#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Анализ лицевых характеристик человека в маске
"""

# ######################################################################################################################
# Импорт необходимых инструментов
# ######################################################################################################################
import os             # Работа с файловой системой
import cv2            # Алгоритмы компьютерного зрения
import pkg_resources  # Работа с ресурсами внутри пакетов

from datetime import datetime  # Работа со временем
from decimal import Decimal, ROUND_HALF_UP  # Округление по математическому закону

# Персональные
from liberty.samples import play  # Пример воспроизведения фото/видео данных


# ######################################################################################################################
# Сообщения
# ######################################################################################################################
class Messages(play.Run):
    """Класс для сообщений"""

    # ------------------------------------------------------------------------------------------------------------------
    # Конструктор
    # ------------------------------------------------------------------------------------------------------------------

    def __init__(self):
        super().__init__()  # Выполнение конструктора из суперкласса

        self._description = self._('Анализ лицевых характеристик человека в маске')
        self._description_time = '{}{}' + self._description + ' ...{}'


# ######################################################################################################################
# Анализ лицевых характеристик человека в маске
# ######################################################################################################################
class MasksViewer(Messages):
    """Анализ лицевых характеристик человека в маске"""

    # ------------------------------------------------------------------------------------------------------------------
    # Конструктор
    # ------------------------------------------------------------------------------------------------------------------

    def __init__(self):
        super().__init__()  # Выполнение конструктора из суперкласса

        self._dnn = ('tf', 'caffe')  # Модели нейронной сети

        self._path_to_models = pkg_resources.resource_filename('facesdet', 'models')  # Путь к моделям

    # ------------------------------------------------------------------------------------------------------------------
    #  Внешние методы
    # ------------------------------------------------------------------------------------------------------------------

    # Загрузка модели для глубокого обучения в OpenCV
    def load_model_opencv_dnn(self, path_to_model = None, path_to_config_model = None, dnn = 'tf', out = True):
        """
        Загрузка модели для метода Виолы-Джонса в OpenCV
        (str, str, str [, bool]) -> bool
        Аргументы:
            path_to_model        - Путь к модели
            path_to_config_model - Путь к конфигурационному файлу модели
            dnn                  - Модель нейронной сети
            out                  - Печатать процесс выполнения
        Возвращает: True если модель загружена, в обратном случае False
        """

        none = 'DL'  # Замена None

        # Путь к модели по умолчанию
        if path_to_model is None:
            path_to_model = none

        # Путь к конфигурационному файлу модели по умолчанию
        if path_to_config_model is None:
            path_to_config_model = none

        # Проверка аргументов
        if type(path_to_model) is not str or not path_to_model or type(path_to_config_model) is not str \
                or not path_to_config_model or type(dnn) is not str or not dnn or type(out) is not bool:
            # Вывод сообщения
            if out is True:
                self._inv_args(__class__.__name__, self.load_model_opencv_dnn.__name__)

            return False

        # Модель нейронной сети не совпадает с необходимыми
        if dnn not in self._dnn:
            return False

        # Путь к модели по умолчанию
        if path_to_model is none:
            path_to_model = \
                os.path.join(self._path_to_models, self._path_to_files_models['opencv_dnn'][dnn]['path_to_model'])

        return True

        # Путь к конфигурационному файлу модели по умолчанию
        if path_to_config_model is none:
            path_to_config_model = self._path_to_models + '/' \
                                   + self._path_to_files_models['opencv_dnn'][dnn]['path_to_config_model']

        required_extension_model = None  # Необходимое расширения файла модели
        required_extension_config_model = None  # Необходимое расширения конфигурационного файла модели

        # 8-битная квантованная версия с использованием TensorFlow
        if dnn == self._dnn[0]:
            # Необходимое расширения файла модели
            required_extension_model = self._required_extension_models[2]

            # Необходимое расширения конфигурационного файла модели
            required_extension_config_model = self._required_extension_models[1]

        # Версия FP16 оригинальной реализации Caffe
        if dnn == self._dnn[1]:
            # Необходимое расширения файла модели
            required_extension_model = self._required_extension_models[4]

            # Необходимое расширения конфигурационного файла модели
            required_extension_config_model = self._required_extension_models[3]

        # Файл модели не найден
        if self._file_manager.search_file(path_to_model, required_extension_model, False, out) is False:
            return False

        # Конфигурационный файл модели не найден
        if self._file_manager.search_file(path_to_config_model, required_extension_config_model, False,
                                          out) is False:
            return False

        # Вывод сообщения
        if out is True:
            print(self._load_model.format(
                datetime.now().strftime(self._format_time),
                self.packages_functions['opencv_dnn']['ru']
            ))

        try:
            # 8-битная квантованная версия с использованием TensorFlow
            if dnn == self._dnn[0]:
                # Чтение модели нейронной сети в формате TensorFlow
                self._model = cv2.dnn.readNetFromTensorflow(path_to_model, path_to_config_model)

            # Версия FP16 оригинальной реализации Caffe
            if dnn == self._dnn[1]:
                # Чтение модели нейронной сети в формате Caffe
                self._model = cv2.dnn.readNetFromCaffe(path_to_config_model, path_to_model)
        except SystemError:
            # Вывод сообщения
            if out is True:
                print(self._model_not_load.format(self.red, datetime.now().strftime(self._format_time), self.end,
                                                  os.path.basename(path_to_model)))

            return False

        self.title_ru = self.packages_functions['opencv_dnn']['ru']  # Установка названия метода на русском
        self.title_en = self.packages_functions['opencv_dnn']['en']  # Установка названия метода на английском

        return True